'use strict';
var contentScrollbar = {
	WINDOW_SCROLL_DURATION: 800, //-// Not used
	MAX_HEADING_SIZE: 3,
	FILTER_HEADINGS: false,
	MIN_MARKER_Z_INDEX:100001,
	_prefService: null,
	_headingMarkers: null,
	_reservedYPosition: null,
	_resevervedYPositionLength: 28,
	_resizeTimeout: null,
	
	init: function(evt) {
		window.removeEventListener('load', contentScrollbar.init);
		gBrowser.addEventListener('load', contentScrollbar.handleContentDocLoad, true);
		window.addEventListener('resize', contentScrollbar.windowResizeHandlerDirect, true);
		contentScrollbar._prefService = contentScrollbar._getPrefService();
	}, 

	unload: function(evt) {
		gBrowser.removeEventListener('load', contentScrollbar.handleContentDocLoad);
	}, 

	handleContentDocLoad: function(evt) {
		var doc = contentScrollbar._getContentDoc(evt.originalTarget);
		if (doc) {
			var contentWin = gBrowser.getBrowserForDocument(gBrowser.contentDocument);
			if(contentWin.__soc_options == undefined) {
				// Must be new tab
				contentWin.__soc_options = { enabled:contentScrollbar._prefService.getBoolPref('autoenabledonnewtab')};
			}			
			if (doc.head) {
				if (contentWin.__soc_options.enabled) {
					contentScrollbar._createHeadingMarkers(doc);
				}
			}
		}
	},
	
	/*
	Returns the content document of the loaded page.
	*/
	_getContentDoc: function(doc) {		
		if (doc instanceof HTMLDocument) {
			var loc = doc.location.href.trim().toLowerCase();
			var scheme = '';

			var index = loc.indexOf(':');
			if (index > 0) {
				scheme = loc.substring(0, index);
			}
						
			if (scheme == 'http' || scheme == 'https' || scheme=='file') {
				// is this an inner frame?  
				if (doc.defaultView.frameElement) {  
				  // Frame within a tab was loaded.  
				  // Find the root document:  
				  while (doc.defaultView.frameElement) {  
					doc = doc.defaultView.frameElement.ownerDocument;  
				  }  
				}
			}
			else {
				doc = null;
			}
		}
		else {
			doc = null;
		}
		return doc;
	},
	
	/*
	_handleStartup: function() {
		var oldVersion = '';
		var currVersion = '0.1.0';
		
		try {
			oldVersion = this._prefService.getCharPref('version');
		}
		catch(e) {}
		
		if (oldVersion != currVersion) {
			this._prefService.setCharPref('version', currVersion);
			try {
				setTimeout(contentScrollbar._welcome,100,currVersion);
			}
			catch(e) {}
		}
	},
	
	_welcome: function(version) {
		try {
			var url = 'http://www.kashiif.com/firefox-extensions/content-scrollbar/content-scrollbar-welcome/?v='+version;
			openUILinkIn( url, 'tab');
		} 
		catch(e) {}
	},
	*/
	
	_getPrefService: function() {
		var prefService = null;
		try 
		{
			prefService = gPrefService;
		}
		catch(err)
		{
			// gPrefService not available in SeaMonkey
			prefService = Components.classes['@mozilla.org/preferences-service;1'].getService(Components.interfaces.nsIPrefService).getBranch('');
		}
		
		prefService = prefService.getBranch('extensions.contentscrollbar.');
		return prefService;
	},
		
	_createHeadingMarkers: function(doc)	{
		contentScrollbar._removeExistingMarkers(doc);
		
		var status = contentScrollbar._prefService.getCharPref('status').toLowerCase();
		if (status == 'disabled') {
			return;
		}

		if(doc.__soc_processed == undefined) {	
			// insert css file
			var nodeToInsert = doc.createElement('link');
			nodeToInsert.setAttribute('rel', 'stylesheet');
			nodeToInsert.setAttribute('type', 'text/css');
			nodeToInsert.setAttribute('href', 'resource://content-scrollbar/cs-client.css');
			doc.head.appendChild(nodeToInsert);

			doc.__soc_processed = true;
		}
		else {
			//Delete existing heading markers if they already exist
			delete contentScrollbar._headingMarkers;
		}
		
		var targetStyleSheet = null;
		for(var i=0 ; i<doc.styleSheets.length; i++) {
			var s1 = doc.styleSheets[i];
			if (s1.href == 'resource://content-scrollbar/cs-client.css') {
				targetStyleSheet = s1;
				break;
			}
		}
		
		if (targetStyleSheet) {
			window.setTimeout( function() {
			
				var markerBackgroundColor = contentScrollbar._getBackGroundColor();

				var rule = targetStyleSheet.cssRules[0]; // .kashiif_soc_marker
				rule.style.backgroundColor = markerBackgroundColor;
				rule.style.boxShadow = '0px 0px 4px rgba(' + contentScrollbar._prefService.getCharPref('backgroundcolor') + ', 0.7)';
				rule.style.fill = markerBackgroundColor; // to be inherited by .kashiif_soc_marker_arrow
				
				rule = targetStyleSheet.cssRules[1]; //.kashiif_soc_marker:hover
				rule.style.backgroundColor = contentScrollbar._getBackGroundColor('1.0');
				
				if (status == 'minimized') {
					rule = targetStyleSheet.cssRules[3]; // .kashiif_soc_marker_span
					rule.style.display = 'none';
				}

			}, 200);
		}
				
		contentScrollbar._headingMarkers = new contentScrollbar.HeadingMarkers();

		// Check if headings exist on the page and create markers if they do.
		var maxHeadingSize = contentScrollbar._prefService.getIntPref('maxheadingsize');
		for( var i = 1; i <= maxHeadingSize; i++ ) {
			var headingTagName = 'h' + i;
			var headings = doc.getElementsByTagName(headingTagName);
			if(headings.length > 0)
				contentScrollbar._headingMarkers.init(doc, headingTagName, headings );
		}
		
		/*------------------------------------------------
		  Set initial markers position and reposition them if window is resized or document 
		  height changes.
		------------------------------------------------*/
		
		contentScrollbar._headingMarkers.setPosition(doc);
	},
	
	_removeExistingMarkers: function(doc) {
		var nodeList = doc.body.querySelectorAll('.kashiif_soc_marker');
		
		for (var i=0 ; i<nodeList.length; i++) {
			var c = doc.getElementById(nodeList[i].id);
			if (c) { doc.body.removeChild(c); }
		}	
	},
	
	_getBackGroundColor: function(factor) {
		if (!factor) {
			factor = '0.78';
		}
		var bgColor = 'rgba(' + contentScrollbar._prefService.getCharPref('backgroundcolor') + ', ' + factor + ' )' ;
		return bgColor;
	},
	
	setMarkersVisibility: function(doc, isHidden) {
		var nodeList = doc.body.querySelectorAll('.kashiif_soc_marker');

		var visibility = 'visible';
		var opacity = '100';

		if (isHidden) {
			visibility = 'hidden';
			opacity = '0';
		}

		for (var i=0 ; i<nodeList.length; i++) {
			var s = nodeList[i].style;
			s.opacity = opacity;			
			s.visibility = visibility;
		}
	},
	
	windowResizeHandlerDirect: function(evt) {
		//alert('resize ' + evt.target + ' ' + evt.originalTarget.document);
		var doc = contentScrollbar._getContentDoc(evt.originalTarget.document);
		if (doc) {
			//contentScrollbar._targetDoc = doc;
			if (contentScrollbar._resizeTimeout) {
				clearTimeout(contentScrollbar._resizeTimeout);
			}
			contentScrollbar._resizeTimeout = setTimeout( function() { contentScrollbar._headingMarkers.windowResizeHandler(doc); }, 500);
		}
	},
	
	/*------------------------------------------------
	  creates a new canvas element, draws an arrow 
	  pointer, and returns it.
	------------------------------------------------*/
	_createArrow: function (doc, markerBackgroundColor) {
		var xmlns = "http://www.w3.org/2000/svg";
	    var arrowCanvas = doc.createElementNS(xmlns, 'svg');
	    arrowCanvas.setAttributeNS(null, 'width', '6px');
	    arrowCanvas.setAttributeNS(null, 'height', '10px');
	    arrowCanvas.setAttribute('class', 'kashiif_soc_marker_arrow');

	    var polygon = doc.createElementNS(xmlns, 'polygon');
	    polygon.setAttributeNS(null, 'points', '0,0 5,4 5,6 0,10');

	    var path = doc.createElementNS(xmlns, 'path');
	    path.setAttributeNS(null, 'd', 'M5,4 Q7,5 5,6');

	    polygon.appendChild(path);
	    arrowCanvas.appendChild(polygon);

	    return arrowCanvas;
	},
	
	HeadingMarkers: function() {
		var socMarkerIDArray = new Array();
		
		this.init = function(doc, headingTagName, headings) {
		
			var textLength = contentScrollbar._prefService.getCharPref('textlength');
			var markerBackgroundColor = contentScrollbar._getBackGroundColor();
			
			for( var i = 0; i < headings.length; i++ ) {
				var headingText = headings[i].textContent.replace(/^\s*/, '').replace(/\s*$/, '').replace(/\s{2,}/, ' ');
				
				/*------------------------------------------------
				  Filter heading
				------------------------------------------------*/
				var filteredHeading = null;
				if(contentScrollbar.FILTER_HEADINGS == true)
					filteredHeading = headingFilters.filterHeading( headings[ i ] );
				else
					filteredHeading = { markHeading: true, filteredHeadingText: headingText };
				
				var markerText = filteredHeading.filteredHeadingText;
				switch(textLength) {
					case 'firstThreeWords':
						var lastIndex = 0;
						var appendDots = '';
						var count=0;
						for (; count<3; count++) {
							var temp = lastIndex;
							lastIndex = markerText.indexOf(' ', lastIndex+1);
							if (lastIndex == -1) {
								lastIndex = markerText.length;
								break;								
							}
						}
						
						if (lastIndex < markerText.length) {
							appendDots = '...';
						}
						
						markerText = markerText.substring(0, lastIndex) + appendDots;
						break;
					
					case 'firstTenCharacters':
						if(markerText.length > 10)
							markerText = markerText.substr(0, 10) + '...';
						break;
				}
				
				if( filteredHeading.markHeading == true ) {
					var markerID = 'soc_' + headingTagName + '_' + ( i + 1 );
					//-*-//var headingY = contentScrollbar._headingMarkers.findPosition( headings[ i ] ).topPos;
					
					/*------------------------------------------------
					  Store each marker's info
					------------------------------------------------*/
					
					socMarkerIDArray.push( markerID );
					
					contentScrollbar._headingMarkers[ markerID ] = { domObject: headings[ i ],
													markerID: markerID,
													tagName: headingTagName,
													headingText: filteredHeading.filteredHeadingText,
													markerText: markerText,
													targetY: 0
													};
					
					/*------------------------------------------------
					  Create a new marker
					------------------------------------------------*/
					
					var newMarker = contentScrollbar._headingMarkers.create(doc, markerID );
					var markerArrow = contentScrollbar._createArrow(doc, markerBackgroundColor);
					
					newMarker.appendChild( markerArrow );
										
					/*------------------------------------------------
					  Add new marker to document
					------------------------------------------------*/
					
					doc.body.appendChild( newMarker );
				}
			}
		}
		
		this.create = function(doc, markerID) {
			var currentHeading = contentScrollbar._headingMarkers[ markerID ];
			var newMarker = doc.createElement( 'div' );
			newMarker.setAttribute( 'id', markerID );
			newMarker.setAttribute( 'class', 'kashiif_soc_marker' );
			
			var spanElement = doc.createElement('span');
			spanElement.setAttribute( 'class', 'kashiif_soc_marker_span' );
			spanElement.textContent = currentHeading.markerText;

			newMarker.appendChild(spanElement);
			
			//-//if( socMarkerDisplay == "minimized" )
				//-//$(newMarker).find(".soc_marker_span").css( { "display": "none" } );
			
			var headingHierarchy = currentHeading.tagName.split('h')[ 1 ];
			var newMarkerZIndex = contentScrollbar.MIN_MARKER_Z_INDEX + contentScrollbar._prefService.getIntPref('maxheadingsize') - headingHierarchy;
			
			//-// newMarker.style = 'z-index:' + newMarkerZIndex + ';display:none;';
			newMarker.style.zIndex = newMarkerZIndex;
			
			
			// Store new marker z-index
			currentHeading.zIndex = newMarkerZIndex;
			
			return newMarker;
		}
		
		/*------------------------------------------------
		  findPosition()
		  
		  Finds an html element's position relative to 
		  the document.
		------------------------------------------------*/
		
		this.findPosition = function( obj ) {
			var leftPos, topPos;
			leftPos = topPos = 0;
			
			do {
				leftPos += obj.offsetLeft;
				topPos += obj.offsetTop;
			}
			while (obj = obj.offsetParent);
			
			return { 'leftPos': leftPos, 'topPos': topPos };
		}
		
		this.setPosition = function(doc) {
			// Components.util.reportError('setPosition: ' + doc);

			var windowHeight = doc.defaultView.innerHeight; //window.innerHeight;
			var winToDocHeightRatio = windowHeight / doc.documentElement.scrollHeight;
			var markerNum = socMarkerIDArray.length;
			
			/*
			alert(  'doc.body.clientHeight: ' + doc.body.clientHeight
				  + '\ndoc.body.scrollHeight: ' + doc.body.scrollHeight
				  + '\nwindow.innerHeight: ' + window.innerHeight 
				  + '\ndoc.defaultView.innerHeight: ' + doc.defaultView.innerHeight
				  + '\nwinToDocHeightRatio doc element: ' + windowHeight / doc.documentElement.scrollHeight
				  + '\nwinToDocHeightRatio body element: ' + windowHeight / doc.body.scrollHeight);
			//*/

			var preventOverlap = contentScrollbar._prefService.getBoolPref('preventoverlap');
			
			if( preventOverlap )
				contentScrollbar._reservedYPosition = new Array();
			
			for( var i = 0; i < markerNum; i++ ) {
				var markerID = socMarkerIDArray[ i ];
				
				/*------------------------------------------------
				  Update each marker's targetY
				------------------------------------------------*/
				var currentHeadingMarker = contentScrollbar._headingMarkers[ markerID ];
				var newTargetY = contentScrollbar._headingMarkers.findPosition( currentHeadingMarker.domObject ).topPos;
				
				currentHeadingMarker.targetY = newTargetY;
				/*------------------------------------------------
				  Reserve y positions then place each marker
				------------------------------------------------*/
				
				var markerY = Number( ( winToDocHeightRatio * currentHeadingMarker.targetY ).toFixed() );
				//var markerY = currentHeadingMarker.targetY;
				var preventOverlap = contentScrollbar._prefService.getBoolPref('preventoverlap');
				if( preventOverlap ) {
					// Check if y position is reserved
					for( var j = 0; j < contentScrollbar._reservedYPosition.length; j++ )
					{
						if( markerY == contentScrollbar._reservedYPosition[ j ] )
							markerY = markerY + 1;
					}
					
					currentHeadingMarker.reservedY = new Array();
					
					for( var k = 0; k < contentScrollbar._resevervedYPositionLength; k++ )
					{
						currentHeadingMarker.reservedY.push( markerY + k );
						contentScrollbar._reservedYPosition.push( markerY + k );
					}					
				}
				
				//alert(markerID + ', newTargetY: ' + newTargetY + ', markerY:'+markerY + ', winToDocHeightRatio: '+ winToDocHeightRatio);
				var markerObject = doc.getElementById(markerID);
				markerObject.style.top = markerY + 'px';
				markerObject.setAttribute('data-newTargetY', newTargetY);
				markerObject.addEventListener('click', function(evt) { 
						var newEvt = this.ownerDocument.createEvent('Events');  
						newEvt.initEvent('contentScrollbar_MarkerClicked', true, true);  
						this.dispatchEvent(newEvt);
					}, false );
			}
		}
		
		this.windowResizeHandler= function(doc) {
			contentScrollbar._headingMarkers.setPosition(doc);
			contentScrollbar._resizeTimeout = null;
		}		
	
		this.getMarkerIds = function() {
			return socMarkerIDArray;
		}
	},

	handleToolbarButtonClick: function(evt) {
		var win = gBrowser.getBrowserForDocument(gBrowser.contentDocument);
		win.__soc_options.enabled = !win.__soc_options.enabled;

		if (win.__soc_options.enabled) {
			if (gBrowser.contentDocument.__soc_processed) {
				// markers already present, simply show them
				contentScrollbar.setMarkersVisibility(gBrowser.contentDocument, false);
			}
			else {
				contentScrollbar._createHeadingMarkers(gBrowser.contentDocument);
			}
		}
		else {
			contentScrollbar.setMarkersVisibility(gBrowser.contentDocument, true);
		}
	},
	
	HeadingFilters: function(){
		this.filterHeading = function(currentHeading) {
			var headingText = currentHeading.innerText.toLowerCase();
			var locationHost = location.host;
			var markHeading, filteredHeadingText;
						
			if(headingText == "")	{
				// Don't create marker for empty headings
						
				markHeading = false;
			}
			else if(locationHost.indexOf('wikipedia.org') != -1)
			{
				// Remove "[Edit]" and its equivalent in other languages in wikipedia headings
						
				var newHeadingText = headingText.replace( /\[\w*\]/, '' );
				filteredHeadingText = newHeadingText;
				
				markHeading = true;
			}
			else if(locationHost.indexOf( 'yahoo' ) != -1)
			{
				var attrClass = currentHeading.getAttribute( 'class' );
				if(attrClass)
				{
					if( attrClass.indexOf('oops') != -1)
						markHeading = false;
					else if( attrClass.indexOf('y-ftr-txt-hdr') != -1)
						markHeading = false;
					else if( attrClass.indexOf('title-date') != -1 )
						markHeading = false;
					else if( attrClass.indexOf('y-txt-1') != -1 && attrClass.indexOf('title') != -1 )
						markHeading = false;
					else if( attrClass.indexOf('y-txt-5') != -1 && attrClass.indexOf('alt-title') != -1 )
						markHeading = false;
					/* replaced by a more specific condition below
					else if( attrClass.indexOf('y-txt-modhdr') )
						markHeading = false;*/
					else if( attrClass.indexOf('y-txt-modhdr') != -1 && currentHeading.getElementsByTagName( "div" )[ 0 ] != undefined )
						markHeading = false;
					else if( currentHeading.getAttribute('id') )
					{
						if( currentHeading.getAttribute('id').indexOf('u_2588582-p-lbl') != -1 )
							markHeading = false;
						else
						{
							filteredHeadingText = headingText;
							markHeading = true;
						}
					}
					else
					{
						filteredHeadingText = headingText;
						markHeading = true;
					}
				}
				else
				{
					filteredHeadingText = headingText;
					markHeading = true;
				}
			}
			else if(locationHost.indexOf('wikia.com') != -1)
			{
				// Remove "Edit" in wikia.com's wikis headings
						
				var newHeadingText = headingText.replace( /edit/i, '');
				filteredHeadingText = newHeadingText;
				
				markHeading = true;
			}
			else
			{	
				filteredHeadingText = headingText;
				markHeading = true;
			}
			
			return { markHeading: markHeading, filteredHeadingText: filteredHeadingText };
		}
	},

	handleMarkerClicked: function(evt) {
		//window.alert('chrome-code: ' + evt.target.getAttribute('data-newTargetY'));
		this.clientWindowScroller.beginScroll(evt.target.getAttribute('data-newTargetY'), evt.target.ownerDocument.defaultView);
	},


	clientWindowScroller : {
	  step: 30,
	  interval: 60,
	  activeScrollTarget: null,
	  activeTimeout: null,
	  beginScroll: function(scrollTarget, clientWin) {
		if (this.activeScrollTarget != scrollTarget) {
			window.clearTimeout(this.activeTimeout);
		  	this.activeScrollTarget = scrollTarget;
		}
		this.animatedScroll(scrollTarget, this.step, clientWin);
	  },
	  animatedScroll: function (scrollTarget, theStep, clientWin) {
		var t = null;
	    var scrollAmount = scrollTarget - clientWin.pageYOffset;
		
		if (scrollAmount>0) {
			if (clientWin.pageYOffset + clientWin.innerHeight >= clientWin.document.documentElement.scrollHeight) {
				// scroll has reached at its maximum value. No More Scrolling possible
				this.activeTimeout = t;
				return;
			}
		}

		
	    var scrollAmountAbs = Math.abs(scrollAmount);
	    if (scrollAmountAbs>theStep) {
		  scrollAmount = (scrollAmount<0?-theStep:theStep);
		  t = window.setTimeout( function(e) { contentScrollbar.clientWindowScroller.animatedScroll(scrollTarget,theStep+10, clientWin); },this.interval);
	    }
	    clientWin.scrollBy(0,scrollAmount);
		this.activeTimeout = t;
	  }
}	
	
};

/*
gBrowser.addEventListener
(
  'load', 
  function (e) { contentScrollbar.init(e); },
  true
);
*/

window.addEventListener
(
  'load', 
  contentScrollbar.init,
  true
);


// Reference: https://developer.mozilla.org/en/Code_snippets/Interaction_between_privileged_and_non-privileged_pages
// The last value is a Mozilla-specific value to indicate untrusted content is allowed to trigger the event.  
window.addEventListener('contentScrollbar_MarkerClicked', function(e) { contentScrollbar.handleMarkerClicked(e); }, false, true); 

//*/
contentScrollbar._headingFilters = new contentScrollbar.HeadingFilters();
